package inf1010.pusle.chat.client;

import inf1010.pusle.chat.*;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Container;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.IOException;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.JList;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleContext;
import javax.swing.text.StyleConstants;
import javax.swing.text.AttributeSet;
import java.util.Vector;

public class ChatFrame extends JFrame implements ChatListener {

    private Vector<String> users;   // Holds the list of users
    private JList userList;         // Displays the list of users
    private JTextPane chatField;    // Displays sent chat messages
    private JTextField inputField;  // Messages are typed into this area
    private ChatClient client;

    private static final Color
        MSG_COLOR = new Color(48, 64, 80),
        JOIN_COLOR = new Color(64, 160, 64),
        LEFT_COLOR = new Color(160, 64, 64),
        USER_COLOR = new Color(80, 48, 32);

    ChatFrame(ChatClient cc) {
        super("PusleChat");

        client = cc;
        client.addChatListener(this);
        Container pane = this.getContentPane();

        users = new Vector<String>();
        userList = new JList(users);
        userList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        userList.setLayoutOrientation(JList.VERTICAL);
        userList.setForeground(MSG_COLOR);

        JScrollPane listScroller = new JScrollPane(userList);
        listScroller.setPreferredSize(new Dimension(100, 0));

        inputField = new JTextField();
        inputField.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent evt) {
                    String message = inputField.getText();
                    if (!message.isEmpty()) {
                        //printMessage(message);
                        client.sendMessage(message);
                        inputField.setText("");
                    }
                }
            });

        chatField = new JTextPane();
        JScrollPane chatScroller = new JScrollPane(chatField);
        chatField.setEditable(false);

        pane.add(listScroller, BorderLayout.LINE_END);
        pane.add(inputField, BorderLayout.PAGE_END);
        pane.add(chatScroller);

        setSize(600, 480);
        setLocationRelativeTo(null);  // Center on screen
        setResizable(true);
        setVisible(true);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        inputField.requestFocus();
    }

    /* Add user s to the user list */
    protected void addUser(String s) {
        users.add(s);
        userList.setListData(users);
    }

    /* Remove user s from the user list */
    protected void removeUser(String s) {
        users.remove(s);
        userList.setListData(users);
    }

    /**
     * Prints a message to the chat window.
     *
     * @param s the message to print.
     */
    protected void printMessage(String s) {
        appendMessage(s + "\n", MSG_COLOR);
    }

    /**
     * Handles incoming packets of certain types.
     *
     * @param m the packet to handle.
     * @throws IOException never
     * @return true if the packet was of a type the ChatFrame can handle.
     */
    public boolean handleMessage(final Message m) throws IOException {
        switch (m.getType()) {
        case ChatUtilities.MESSAGE_FROM:
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        appendMessage(m.getData() + ": ", USER_COLOR);
                    }
                });
            return true;

        case ChatUtilities.MESSAGE:
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        printMessage(m.getData());
                    }
                });
            return true;

        case ChatUtilities.NEW_USER:
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        appendMessage(String.format("*** %s joined the chat ***%n",
                                                   m.getData()), JOIN_COLOR);
                        addUser(m.getData());
                    }
                });
            return true;

        case ChatUtilities.USER_LEFT:
            SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        appendMessage(String.format("*** %s left the chat ***%n",
                                                   m.getData()), LEFT_COLOR);
                        removeUser(m.getData());
                    }
                });
            return true;

        case ChatUtilities.USER_LIST:
            if (m.getData() != null && !m.getData().isEmpty()) {
                SwingUtilities.invokeLater(new Runnable() {
                        public void run() {
                            addUser(m.getData());
                        }
                    });
            }
            return true;
        }

        return false;
    }

    /* Append message s to the message field in color c */
    private void appendMessage(String s, Color c) {
        StyleContext sc = StyleContext.getDefaultStyleContext();
        AttributeSet attr = sc.addAttribute(SimpleAttributeSet.EMPTY,
                                            StyleConstants.Foreground, c);
        int len = chatField.getDocument().getLength();
        chatField.setCaretPosition(len);
        chatField.setCharacterAttributes(attr, false);
        chatField.setEditable(true);
        chatField.replaceSelection(s);
        chatField.setEditable(false);
    }
}