package inf1010.pusle.chat.server;

import inf1010.pusle.chat.*;

import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.net.*;
import java.io.*;

/**
 * A simple chat server capable of serving multiple clients.
 */
public class ChatServer implements Runnable {

    /** The default port used by the server if no port is defined. */
    private static final int DEFAULT_PORT = 9513;

    /** 60 second timeout for login. */
    private static final long LOGIN_TIMEOUT = 60000000;

    /** The socket used to accept connections. */
    private ServerSocket ss;

    /** Logged in clients. */
    private Map<String, Client> clients = new HashMap<String, Client>();

    /** Clients awaiting login. */
    private Map<Long, Client> clientQueue = new HashMap<Long, Client>();

    /**
     * Entry point.
     *
     * @args The first argument is the port.
     */
    public static void main(String[] args) {
        int port = DEFAULT_PORT;
        if (args.length > 0) {
            try {
                port = Integer.parseInt(args[0]);
            } catch (NumberFormatException nfe) {
                System.err.println("First argument should be a port number.");
                System.err.println(nfe.getMessage());
                System.exit(1);
            }
        }

        ChatServer cs = new ChatServer(port);
        cs.run();
    }

    /**
     * Creates a ChatServer binding a ServerSocket to the given port.
     * 
     * @param port the port to bind the server socket to.
     */
    public ChatServer(int port) {
        try {
            // Create and bind a socket to the given port
            ss = new ServerSocket(port);
            // Set the timeout for calls to accept
            // (accept will block for 1 second)
            ss.setSoTimeout(1000);
        } catch (IOException bindException) {
            System.out.printf("Could not bind socket on port #%d.%n", port);
            System.exit(2);
        }

        // Handle shutdowns "gracefully"
        // This might give you some hints as to what to do elsewhere
        Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    Iterator<Client> it = clients.values().iterator();
                    while(it.hasNext()) {
                        it.next().disconnect("Shutting down.", true);
                        it.remove();
                    }
                    System.out.println("Good bye!");
                    try {
                        ss.close();
                    } catch(IOException e) {
                        System.err.println("Could not close server socket.");
                        System.err.println(e.getMessage());
                    }
                }
            });
    }

    /**
     * Processes the queue of connecting clients.
     */
    private void processQueue() {
        Iterator<Map.Entry<Long, Client>> it = clientQueue.entrySet().iterator();
        long currentTime = System.nanoTime();

        while (it.hasNext()) {
            Map.Entry<Long, Client> entry = it.next();
            long oldTime = entry.getKey();
            Client c = entry.getValue();
            String name = c.getName();

            if (name != null) {
                if (name.isEmpty()) {
                    // INVALID_PACKET
                    /* TODO Remove client who sent an invalid packet. */
                    continue;
                } else if (clients.containsKey(name)) {
                    // INVALID_USERNAME
                    /*
                     * TODO Remove client who sent a taken user name.
                     * Remember to call dename,
                     * so disconnecting the user doesn't remove the user with the same name.
                     */
                    continue;
                }

                // WELCOME
                // Login successful - remove from queue of clients waiting to log in
                it.remove();

                /*
                 * TODO Send welcome message and user list to user
                 * Tell all clients that the user has joined the chat
                 */
                
            } else if (currentTime - oldTime > LOGIN_TIMEOUT) {
                /* TODO disconnect client who spent too long before logging in. */
            }
        }
    }

    /**
     * Disconnects a client and removes it from the list of clients.
     *
     * @param c the client to remove.
     */
    public synchronized void removeClient(Client c) {
        String name = c.getName();

        if(clients.remove(name) == null)
	    return;

        /*
         * TODO Disconnect the client and tell other users that the client left
         */

    }

    /**
     * Sends a chat message on the network.
     *
     * @param msg the message to be sent.
     * @param sender the client sending the message.
     */
    public synchronized void broadcast(String msg, Client sender) {
        /*
         * TODO Send the chat message from sender to all users.
         */
    }

    /**
     * Wait for client connections and add them to the list of clients.
     */
    public void run() {
        while (true) {
            try {
                processQueue();

                /*
                 * TODO accept client connections and start a thread for the client
                 */

            } catch (SocketTimeoutException ste) {
                // DO NOTHING, we want to process the queue once in a while
            } catch (IOException ioe) {
                System.err.println("Could not accept connection.");
                System.err.println(ioe.getMessage());
            }
        }
    }
}