package inf1010.pusle.chat.server;

import inf1010.pusle.chat.*;

import java.io.*;
import java.net.*;

/**
 * Represents a client connected to a server.
 */
class Client implements Runnable {

    /** The chat server this client belongs to. */
    private ChatServer server;

    /** The socket this client is connected on. */
    private Socket s;

    /** The socket's input stream. */
    private InputStream sin;

    /** The socket's output stream. */
    private OutputStream sout;

    /** The user's name. */
    private String name;

    /**
     * Creates a new client with the associated socket.
     *
     * @param s the socket to associate with this client.
     * @throws IOException if opening the input and output streams of the socket fails.
     */
    public Client(Socket s, ChatServer server) throws IOException {
        this.s = s;
        this.sin = s.getInputStream();
        this.sout = s.getOutputStream();
        this.server = server;
    }

    /**
     * Disconnects a client from the server.
     *
     * @param msg the message to send the client when disconnecting it.
     * @param serverSide whether the disconnect is delegated by the server.
     */
    public synchronized void disconnect(String msg, boolean serverSide) {
        try {
            if (serverSide) {
                ChatUtilities.sendMessage(ChatUtilities.DISCONNECT, msg, sout);
            }
            s.close();
        } catch (IOException e) {
            System.err.println("Error occurred while disconnecting user:");
            System.err.println(e.getMessage());
        }
    }

    /**
     * Handles incoming packets from a client.
     *
     * @param m the packet to handle.
     * @throws IOException if sending error messages to the user fails.
     */
    public void handleMessage(Message m) throws IOException {
        if (name == null) {
            login(m);
            return;
        }

        switch (m.getType()) {
        case ChatUtilities.MESSAGE:
            server.broadcast(m.getData(), this);
            break;
        case ChatUtilities.DISCONNECT:
            server.removeClient(this);
            break;
        default:
            ChatUtilities.sendMessage(ChatUtilities.INVALID_PACKET,
                                      String.format("%s%n%s",
                                                    "Unexpected packet type.",
                                                    "Expected MESSAGE or DISCONNECT."),
                                      sout);
        }
    }

    /**
     * Logs in a user if they send a login packet, otherwise they get thrown out.
     *
     * @param m the packet that was received from the client.
     */
    public void login(Message m) {
        if (m.getType() != ChatUtilities.LOGIN) {
            try {
                ChatUtilities.sendMessage(ChatUtilities.INVALID_PACKET,
                                          "You must log in first.",
                                          sout);
            } catch (IOException ioe) {
                System.err.println("Could not send error to client not logging in.");
                System.err.println(ioe.getMessage());
            }
            name = "";
            return;
        }

        name = m.getData();
        if (name == null) {
            name = "";
        }
    }

    /**
     * Unsets this client's name.
     */
    public void dename() {
	name = null;
    }

    /**
     * Get incoming packets.
     */
    public void run() {
        while (!s.isClosed()) {
            try {
                Message m = ChatUtilities.recvMessage(sin);
                handleMessage(m);
            } catch (IOException e) {
                System.err.printf("Client error (%s):%n", name);
                System.err.println(e.getMessage());
                server.removeClient(this);
            }
        }
    }

    /**
     * Sends a message using
     * {@link ChatUtilities#sendMessage(byte, String, OutputStream)}.
     *
     * @param type the type of packet to send.
     * @param data the data to send.
     * @throws IOException if anything goes wrong with sending the data.
     * @return false if the data was too long.
     * @see ChatUtilities#sendMessage(byte, String, OutputStream)
     */
    public boolean sendMessage(byte type, String data) throws IOException {
        return ChatUtilities.sendMessage(type, data, sout);
    }

    /**
     * Gets the username for this client.
     *
     * @return the username for this client.
     */
    public String getName() {
        return name;
    }
}